/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"
#include <stdio.h>

#include <Xm/DialogS.h>
#include <Xm/ScrolledW.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/TextF.h>
#include <Xm/Text.h>
#include <Xm/Scale.h>
#include <Xm/Label.h>
#include <Xm/Frame.h>
#include <Xm/BulletinB.h>
#include <Xm/SelectioB.h>
#include <Xm/DrawnB.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include "xmap.h"
#include "gps.h"
#include "mode.h"
#include "debug.h"

static Widget dialog = NULL;
static Widget bul;
static Widget cancel;
static Widget b1, b2, b3, b4, b5, b6;
static Widget text;
static Widget scale;

extern XtAppContext app;
extern Widget toplevel;

static void read_next();

static struct header {
    int major;
    int minor;
    char wday[10];
    char month[10];
    int day;
    char time[10];
    char year[10];
    struct tm tm;
} h;

static struct tm ct;
static struct stat st;

static int answer = 0;
static int warp = 10;
static int cf = 0;
static int playing = 0;
static long fsize;
static long fpos;
static FILE *ifp;

static void update_scale()
{
	double s, p, val;

	fpos = ftell(ifp);
	s = (double) fsize;
	p = (double) fpos;

	val = p / s * 100.0;

	XtVaSetValues(scale,
		XmNvalue, (int) val,
		NULL);
}

static void update_text(char *bufp)
{
	if (bufp[strlen(bufp)-1] == '\n') bufp[strlen(bufp)-1] = '\0';
	XtVaSetValues(text,
		XmNvalue, bufp,
		NULL);
}

static int find_section(FILE * fp)
{
    char bufp[256];

    while (fgets(bufp, 255, fp) != NULL) {
	update_scale();
	if (strncmp("hugolog", bufp, 7) == 0) {
	    sscanf(bufp, "hugolog %d.%d %s %s %d %s %s", &h.major, &h.minor,
		   h.wday, h.month, &h.day, h.time, h.year);

	    snprintf(bufp, 255, "%s %d %s %s", h.month, h.day, h.year, h.time);
	    strptime(bufp, "%b %e %Y %T", &h.tm);
	    snprintf(bufp, 255, "Section: %s", asctime(&h.tm));
	    update_text(bufp);
	    return 1;
	}
    }
    update_scale();
    return 0;
}

static void ff_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
	if (!find_section(ifp)) beep();
}

static void rew_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
	rewind(ifp);
	find_section(ifp);
}

static void cancel_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    answer = 1;
    gpsdata.status = -1;	//Or not?

    mark_track();
    playing = 0;
    ch_mode(TRACKFROMDISK,STOP);
}

static void stop_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    gpsdata.status = -1;	//Or not?
    playing = 0;
}

static void warp_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    XmString s;
    char bufp[256];

    warp *= 10;

    if (warp > 100)
	warp = 1;

    snprintf(bufp, 255, "x%d", warp);
    s = XmStringCreateSimple(bufp);
    XtVaSetValues(b1,
		  XmNlabelString, s,
		  NULL);
    XmStringFree(s);
}

static void segment_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{

    cf ^= 1;

    XtVaSetValues(b2,
		  XmNlabelPixmap, pbutton[cf].pmap,
		  NULL);
}

static void plott()
{
    mark_track();
    show_log_status();
    if (playing) read_next();
}

static void read_next()
{
    double dt;
    char bufp[256];

    if (fgets(bufp, 255, ifp) != NULL) {
    	update_scale();
	if (strncmp("end", bufp, 3) == 0) {
	    sscanf(bufp, "end %s %s %d %s %s",
		   h.wday, h.month, &h.day, h.time, h.year);
	    beep();
	    if (cf) playing = 0;
	    else XtAppAddTimeOut(app, 10L, (XtTimerCallbackProc) plott, NULL);
	    return;
	} 

	else if (strncmp("hugolog", bufp, 7) == 0) {
	    sscanf(bufp, "hugolog %d.%d %s %s %d %s %s", &h.major, &h.minor,
		   h.wday, h.month, &h.day, h.time, h.year);

	    snprintf(bufp, 255, "%s %d %s %s", h.month, h.day, h.year, h.time);
	    strptime(bufp, "%b %e %Y %T", &h.tm);
	    snprintf(bufp, 255, "Section: %s", asctime(&h.tm));
	    update_text(bufp);
	    XtAppAddTimeOut(app, 10L, (XtTimerCallbackProc) plott, NULL);
	    return;
	}

	else {		// Must be valid record

	    gpsdata.mode = 0;
	    sscanf(bufp, "%d %lf %lf %lf %lf %s %s",
		   &gpsdata.status, &gpsdata.latitude, &gpsdata.longitude,
		   &gpsdata.velocity, &gpsdata.altitude,
		   gpsdata.date, gpsdata.time);
	    snprintf(bufp, 255, "%s %s", gpsdata.time, gpsdata.date);
	    strptime(bufp, "%H:%M:%S %m/%d/%Y", &ct);
	    if (playing == 1) {
		playing++;
		dt = 0.0;
	    } else {
		dt = difftime(mktime(&ct), mktime(&h.tm));
		dt *= 1000;
		dt /= (double) warp;
	    }
	    memcpy(&h.tm, &ct, sizeof(ct));

	    /*FIXME*/
	    // This is an ugly one!
	    // For some reason strptime gives sometimes strange results.
	    if (dt > 1000.0)
		dt = 100.0;
	    if (dt < 0.0)
		dt = 100;
	    XtAppAddTimeOut(app, (long) dt, (XtTimerCallbackProc) plott, NULL);
	}
    }
    else playing = 0;
}

static void play_cb(Widget w, XtPointer client_data, XmAnyCallbackStruct * cbs)
{
    if (!playing) {
	playing = 1;
	read_next();
    } else
	beep();
}


#define H 80

void replay(char *name, FILE * fp)
{
    Pixel foreground;
    XmString s;
    char bufp[256];

    ch_mode(TRACKFROMDISK,START);
    if (!dialog) {
	s = XmStringCreateSimple("");
	foreground = get_pixel(toplevel, "Black");
	dialog = XtVaCreatePopupShell("playdialog",
			 xmDialogShellWidgetClass, GetTopShell(toplevel),
				      XmNdeleteResponse, XmUNMAP,
				      XmNresizePolicy, XmRESIZE_NONE,
				      XmNnoResize, True,
				      XmNwidth, 300,
				      XmNminWidth, 300,
				      XmNmaxWidth, 300,
				      XmNheight, H,
				      XmNminHeight, H,
				      XmNmaxHeight, H,
				      XmNmarginWidth, 5,
				      XmNmarginHeight, 5,
				      NULL);

	bul = XtVaCreateWidget("bul",
			       xmBulletinBoardWidgetClass, dialog,
			       XmNmarginHeight, 6,
			       XmNforeground, foreground,
			       XmNautoUnmanage, False,
			       XmNdialogStyle, XmDIALOG_WORK_AREA,
			       NULL);
	XtAddCallback(bul, XmNunmapCallback, (void *) cancel_cb, NULL);


	text = XtVaCreateManagedWidget("text",
				xmTextFieldWidgetClass, bul,
				XmNx, 20,
				XmNy, 6,
				XmNwidth, 260,
				XmNheight, 20,
				XmNhighlightThickness, 0,
				XmNshadowThickness, 1,
				XmNeditable, False,
				XmNcursorPositionVisible, False,
				XmNmarginHeight, 2,
				XmNmarginWidth, 3,
				NULL);

	scale = XtVaCreateManagedWidget("scale",
				xmScaleWidgetClass, bul,
				XmNx, 20,
				XmNy, 30,
				XmNorientation, XmHORIZONTAL,
				XmNprocessingDirection, XmMAX_ON_RIGHT,
				XmNwidth, 260,
				XmNheight, 16,
				XmNhighlightThickness, 0,
				NULL);


	cancel = XtVaCreateManagedWidget("Cancel",
					 xmPushButtonWidgetClass, bul,
					 XmNforeground, foreground,
					 XmNhighlightThickness, 0,
					 XmNwidth, 60,
					 XmNx, 210,
					 XmNy, H - 28,
					 NULL);
	XtAddCallback(cancel, XmNactivateCallback, (void *) cancel_cb, NULL);

	b1 = XtVaCreateManagedWidget("warp",
				     xmDrawnButtonWidgetClass, bul,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 35,
				     XmNheight, 24,
				     XmNx, 20,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b1, XmNactivateCallback, (void *) warp_cb, NULL);

	b2 = XtVaCreateManagedWidget("cont",
				     xmDrawnButtonWidgetClass, bul,
				     XmNlabelType, XmPIXMAP,
				     XmNlabelPixmap, pbutton[cf].pmap,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 20,
				     XmNx, 65,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b2, XmNactivateCallback, (void *) segment_cb, NULL);

	b3 = XtVaCreateManagedWidget("rew",
				     xmDrawnButtonWidgetClass, bul,
				     XmNlabelType, XmPIXMAP,
				     XmNlabelPixmap, pbutton[2].pmap,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 20,
				     XmNx, 90,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b3, XmNactivateCallback, (void *) rew_cb, NULL);

	b4 = XtVaCreateManagedWidget("stop",
				     xmDrawnButtonWidgetClass, bul,
				     XmNlabelType, XmPIXMAP,
				     XmNlabelPixmap, pbutton[3].pmap,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 20,
				     XmNx, 115,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b4, XmNactivateCallback, (void *) stop_cb, NULL);

	b5 = XtVaCreateManagedWidget("play",
				     xmDrawnButtonWidgetClass, bul,
				     XmNlabelType, XmPIXMAP,
				     XmNlabelPixmap, pbutton[4].pmap,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 20,
				     XmNx, 140,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b5, XmNactivateCallback, (void *) play_cb, NULL);

	b6 = XtVaCreateManagedWidget("ff",
				     xmDrawnButtonWidgetClass, bul,
				     XmNlabelType, XmPIXMAP,
				     XmNlabelPixmap, pbutton[5].pmap,
				     XmNforeground, foreground,
				     XmNhighlightThickness, 0,
				     XmNwidth, 20,
				     XmNx, 165,
				     XmNy, H - 30,
				     NULL);
	XtAddCallback(b6, XmNactivateCallback, (void *) ff_cb, NULL);

	XmStringFree(s);

    }
    snprintf(bufp, 255, "x%d", warp);
    s = XmStringCreateSimple(bufp);
    XtVaSetValues(b1,
		  XmNlabelString, s,
		  NULL);
    XmStringFree(s);


    XtVaSetValues(dialog,
		  XmNtitle, name,
		  NULL);


    ifp = fp;


    if (fstat(fileno(ifp), &st)) {
	error("Could not get information about this file.\n",errno);
	return;
    }

    fsize = st.st_size;
    update_scale();


    if (!find_section(fp)) {
	error("This file does not contain valid data.\nPossibly not a hugo log file.", 0);
	return;
    }

    answer = 0;
    XtManageChild(bul);
    XtManageChild(dialog);
    XtPopup(dialog, XtGrabNone);

    while (answer == 0)
	XtAppProcessEvent(app, XtIMAll);

    XtPopdown(dialog);
    /* make sure the dialog goes away before returning. Sync with server
     * and update the display.
     */
    XSync(XtDisplay(dialog), 0);
    XmUpdateDisplay(XtParent(dialog));
}
